@@ -1,5 +1,7 @@
 /* SPDX-License-Identifier: GPL-2.0 */
 
+#define _GNU_SOURCE
+
 #include <errno.h>
 #include <unistd.h>
 #include <stdlib.h>
@@ -148,6 +150,21 @@ out:
 	return _libbpf_version;
 }
 
+static bool try_bpf_file(char *buf, size_t buf_size, char *path,
+			 const char *progname)
+{
+	struct stat sb = {};
+
+	if (try_snprintf(buf, buf_size, "%s/%s", path, progname))
+		return false;
+
+	pr_debug("Looking for '%s'\n", buf);
+	if (stat(buf, &sb))
+		return false;
+
+	return true;
+}
+
 int find_bpf_file(char *buf, size_t buf_size, const char *progname)
 {
 	static char *bpf_obj_paths[] = {
@@ -157,21 +174,15 @@ int find_bpf_file(char *buf, size_t buf_size, const char *progname)
 		BPF_OBJECT_PATH,
 		NULL
 	};
-	struct stat sb = {};
-	char **path;
-	int err;
-
-	for (path = bpf_obj_paths; *path; path++) {
-		err = try_snprintf(buf, buf_size, "%s/%s", *path, progname);
-		if (err)
-			return err;
-
-		pr_debug("Looking for '%s'\n", buf);
-		err = stat(buf, &sb);
-		if (err)
-			continue;
+	char *path, **p;
 
+	path = secure_getenv(XDP_OBJECT_ENVVAR);
+	if (path && try_bpf_file(buf, buf_size, path, progname)) {
 		return 0;
+	} else if (!path) {
+		for (p = bpf_obj_paths; *p; p++)
+			if (try_bpf_file(buf, buf_size, *p, progname))
+				return 0;
 	}
 
 	pr_warn("Couldn't find a BPF file with name %s\n", progname);
@@ -852,3 +863,21 @@ int iface_print_status(const struct iface *iface)
 out:
 	return err;
 }
+
+int iface_get_xdp_feature_flags(int ifindex, __u64 *feature_flags)
+{
+#ifdef HAVE_LIBBPF_BPF_XDP_QUERY
+	LIBBPF_OPTS(bpf_xdp_query_opts, opts);
+	int err;
+
+	err = bpf_xdp_query(ifindex, 0, &opts);
+	if (err)
+		return err;
+
+	*feature_flags = opts.feature_flags;
+	return 0;
+#else
+	return -EOPNOTSUPP;
+#endif
+
+}
