@@ -31,6 +31,7 @@
 #include <linux/in6.h>
 #include <linux/udp.h>
 #include <linux/tcp.h>
+#include <linux/netdev.h>
 
 #include "params.h"
 #include "logging.h"
@@ -54,6 +55,68 @@ DEFINE_SAMPLE_INIT(xdp_trafficgen);
 
 static bool status_exited = false;
 
+static const char *driver_pass_list[] = {
+	"bnxt",
+	"ena",
+	"gve",
+	"i40e",
+	"ice",
+	"igb",
+	"igc",
+	"ixgbe",
+	"octeontx2",
+	"stmmac",
+};
+
+static bool driver_needs_xdp_pass(const struct iface *iface)
+{
+	const char *name = get_driver_name(iface->ifindex);
+	__u64 feature_flags;
+	size_t i;
+	int err;
+
+	/* If the interface already has the NDO_XMIT feature, we don't need to load anything */
+	err = iface_get_xdp_feature_flags(iface->ifindex, &feature_flags);
+	if (!err && feature_flags & NETDEV_XDP_ACT_NDO_XMIT)
+		return false;
+
+	for (i = 0; i < ARRAY_SIZE(driver_pass_list); i++) {
+		if (!strcmp(name, driver_pass_list[i])) {
+			pr_debug("Driver %s on interface %s needs an xdp_pass program to use XDP_REDIRECT\n",
+				 name, iface->ifname);
+			return true;
+		}
+	}
+
+	return false;
+}
+
+static int check_iface_support(const struct iface *iface)
+{
+	__u64 feature_flags = 0;
+	int err;
+
+	err = iface_get_xdp_feature_flags(iface->ifindex, &feature_flags);
+	if (err || !feature_flags) {
+		/* The libbpf query function, doesn't distinguish between
+		 * "querying is not supported" and "no feature flags are set",
+		 * so treat a 0-value feature_flags as a failure to query
+		 * instead of refuring to run because the NDO_XMIT bit is not
+		 * set.
+		 */
+		pr_warn("Couldn't query XDP features for interface %s (%d).\n"
+			"Continuing anyway, but running may fail!\n",
+			iface->ifname, -err);
+	} else if (!(feature_flags & NETDEV_XDP_ACT_NDO_XMIT)) {
+		pr_warn("Interface %s does not support sending packets via XDP.\n"
+			"Most likely the driver is missing support in this kernel version.\n",
+			iface->ifname);
+		return -EOPNOTSUPP;
+	}
+
+	return 0;
+}
+
 struct udp_packet {
 	struct ethhdr eth;
 	struct ipv6hdr iph;
@@ -400,6 +463,7 @@ static struct prog_option udp_options[] = {
 
 int do_udp(const void *opt, __unused const char *pin_root_path)
 {
+	struct xdp_program *prog = NULL, *pass_prog = NULL;
 	const struct udpopt *cfg = opt;
 
 	DECLARE_LIBXDP_OPTS(xdp_program_opts, opts);
@@ -411,7 +475,6 @@ int do_udp(const void *opt, __unused const char *pin_root_path)
 	struct xdp_trafficgen *skel = NULL;
 	struct udp_packet *payload = NULL;
 	pthread_t *runner_threads = NULL;
-	struct xdp_program *prog = NULL;
 	int err = 0, i;
 	char buf[100];
 	__u32 key = 0;
@@ -459,10 +522,35 @@ int do_udp(const void *opt, __unused const char *pin_root_path)
 		goto out;
 	}
 
+	if (driver_needs_xdp_pass(&cfg->iface)) {
+		opts.prog_name = "xdp_pass";
+		pass_prog = xdp_program__create(&opts);
+		if (!pass_prog) {
+			err = -errno;
+			pr_warn("Couldn't load xdp_pass program\n");
+			goto out;
+		}
+	}
+
 	err = xdp_trafficgen__load(skel);
 	if (err)
 		goto out;
 
+	if (pass_prog) {
+		err = xdp_program__attach(pass_prog, cfg->iface.ifindex,
+					  XDP_MODE_NATIVE, 0);
+		if (err) {
+			pr_warn("Couldn't attach xdp_pass program");
+			xdp_program__close(pass_prog);
+			pass_prog = NULL;
+			goto out;
+		}
+	}
+
+	err = check_iface_support(&cfg->iface);
+	if (err)
+		goto out;
+
 	err = bpf_map_update_elem(bpf_map__fd(skel->maps.state_map),
 				  &key, &bpf_state, BPF_EXIST);
 	if (err) {
@@ -493,6 +581,11 @@ int do_udp(const void *opt, __unused const char *pin_root_path)
 	}
 
 out:
+	if (pass_prog) {
+		xdp_program__detach(pass_prog, cfg->iface.ifindex,
+				    XDP_MODE_NATIVE, 0);
+		xdp_program__close(pass_prog);
+	}
 	xdp_program__close(prog);
 	xdp_trafficgen__destroy(skel);
 	free(runner_threads);
@@ -751,6 +844,10 @@ int do_tcp(const void *opt, __unused const char *pin_root_path)
 	}
 	attached = true;
 
+	err = check_iface_support(&cfg->iface);
+	if (err)
+		goto out;
+
 	err = bpf_map_update_elem(bpf_map__fd(state_map),
 				  &key, &bpf_state, BPF_EXIST);
 
@@ -926,14 +1023,25 @@ out:
 }
 
 static const struct probeopt {
+	struct iface iface;
 } defaults_probe = {};
 
-static struct prog_option probe_options[] = {};
+static struct prog_option probe_options[] = {
+	DEFINE_OPTION("interface", OPT_IFNAME, struct probeopt, iface,
+		      .metavar = "<ifname>",
+		      .short_opt = 'i',
+		      .help = "Probe features of device <ifname>"),
+};
 
-int do_probe(__unused const void *cfg, __unused const char *pin_root_path)
+int do_probe(const void *opt, __unused const char *pin_root_path)
 {
-	int err = probe_kernel_support();
+	const struct probeopt *cfg = opt;
+	int err;
 
+	if (cfg->iface.ifindex)
+		check_iface_support(&cfg->iface);
+
+	err = probe_kernel_support();
 	if (!err)
 		pr_info("Kernel supports live packet mode for XDP BPF_PROG_RUN.\n");
 	return err;
